/* daemon.c ver 1.0  04/03/93 */

/*
 *  Copyright (C) 1993 University of Liverpool.
 */

/* 
 *
 *  Daemon for 5.12 (Berkeley) TFTP server.
 *
 */

#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#if defined(__hpux) || defined(SCOOS5)
#include <fcntl.h>
#include <memory.h>
#include <time.h>
#include <unistd.h>
#else
#include <sys/file.h>
#include <sys/time.h>
#include <sys/wait.h>
#endif
#include <sys/errno.h>
#include <sys/types.h>  
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/tftp.h>

#if defined(BSD)
#define UTSLEN 9
#endif
#if defined(SCOOS5)
#define UTSLEN MAXHOSTNAMELEN
#endif

#ifndef LOGFILE
#define LOGFILE "/usr/adm/tftpd.log"
#endif
#ifndef TFTP_UID
#define TFTP_UID 20
#endif
#define PID     "/etc/tftpd.pid"
#define RETRY 10    /* seconds to wait before retrying after a failed call */

/* Global data */
extern int errno;
extern char *sys_errlist[];

#define MAXARG  4
extern char    *dirs[MAXARG+1];
#define PKTSIZE SEGSIZE+4
extern char buf[PKTSIZE]; 

FILE    *logfile;	     /* status/error logging */ 
int       log_fd;
char hostname[UTSLEN+1];     /* server hostname */
extern struct sockaddr_in from;     /* client address */

/* Functions */
extern char *inet_ntoa();
extern void serve_request();
extern void signal_current_tftpd();
void closedown();
void help();
void write_my_pid();
void signal_current_tftpd();

main(argc, argv)
int argc;
char *argv[];
{
int                     lis,     /* client request socket */
	    lflag=0,sflag=0;     /* command line flags */
u_long              rnum=0L;     /* client request serial number  */
struct hostent          *hp;     /* hosts entry for host and clients */
struct servent          *sp;     /* /etc/services entry for "tftp" port */
struct sockaddr_in myaddr_in;    /* server address */

  /*
   *  Parse command line args.
   */
  {
  register int n=0,i=0;
    while (++n < argc)
      if ((argv[n][0]) == '-')
        switch(argv[n][1])
        {
        case 'h':  help();
        case 'k':  (void)signal_current_tftpd(SIGTERM);   /* shut down */
                            exit(0);
        case 'l':  ++lflag;                               /* log connections */
                   break;
        case 's':  ++sflag;                               /* run as daemon */
                   break;
        default:   fprintf(stderr,"Unknown option: -%c\n",argv[n][1]);
                   exit(1);
        }
      else if (i <= (MAXARG-1))				 /* add a home directory to the list */
	dirs[i++] = argv[n];
  }

  /*
   *  Assume being run from inetd.
   */
  if (!sflag) serve_request(0, 0);
  

  /* get local host name and IP address */
  if (gethostname(hostname, UTSLEN) == -1)
  {
    perror("tftpd: error: don't know my own hostname !");
    exit(1);
  }
  if ((hp=gethostbyname(hostname)) == NULL)
  {
    fprintf(stderr,"tftpd: error: no network address for %s",hostname);
    exit(1);
  }

  /*
   *  Get tftp service port number.
   */
  if ((sp=getservbyname("tftp", "udp")) == NULL)
  {
    fputs("tftpd: No service port for tftp",stderr);
    exit(1);
  }

  /*
   *  Open error/status log file. 
   */
#ifndef DEBUG
  if (!lflag) 
  {
    if ((log_fd=open("/dev/null",  O_WRONLY | O_APPEND, 0)) == -1)
    {
      perror("tftpd: could not open null device ?");
      exit(1);
    }
  }
  else if ((log_fd=open(LOGFILE, O_CREAT | O_WRONLY | O_APPEND, 0600)) == -1)
  {
    perror("tftpd: could not open log file");
    exit(1);
  }
  if ((logfile=fdopen(log_fd, "a")) == NULL)
  {
     perror("tftpd: could not open log file");
     exit(1);
  }
  (void)setvbuf(logfile, NULL, _IOLBF, 0);
#else
  logfile=stdout;
#endif

  /*
   *  Set up connectionless listen socket.
   */
  memset (&myaddr_in, 0, sizeof(struct sockaddr_in)); 
  memset (&from, 0, sizeof(struct sockaddr_in)); 
  myaddr_in.sin_addr.s_addr=INADDR_ANY; /* set network address to wildcard */
  myaddr_in.sin_family=AF_INET;
#ifndef DEBUG
  myaddr_in.sin_port=sp->s_port; 
#else
  myaddr_in.sin_port=TESTPORT;
#endif
  /* open listen socket */
  if ((lis=socket(AF_INET, SOCK_DGRAM, 0)) == -1)
  {
    perror("tftpd: could not open listen socket");
    exit(1);
  }
  /* attempt bind to listen socket */
  if ((bind(lis, &myaddr_in, sizeof(struct sockaddr_in))) == -1)
  {
    perror("tftpd: Unable to bind listen socket: ");
    exit(1);
  }

#ifndef DEBUG
  /*
   *  Background daemon. 
   */
  switch(fork())
  {
    case -1: 
          perror("tftpd: could not fork to background process");
          exit(1);
    case 0:
          {
        int fd;
          (void) close(0);
          (void) close(1);
          (void) close(2);
          (void) open("/dev/null", 0);
          (void) dup2(0, 1);
          (void) dup2(0, 2); 
#ifdef BSD
          if ((fd=open("/dev/tty", O_RDWR)) > 0);
          {
            (void) ioctl(fd, TIOCNOTTY, (char *)0);
            (void) close(fd);
          }
          (void)setpgrp(0, 0);
#else
          (void)setpgrp(); /* Voids terminal assosciation sets pgid = pid */
#endif 
          break;
          }
    default: exit(0);
  }
#endif
  /*
   *  Complete initialisation of daemon. 
   */
#ifdef __hpux
  signal(SIGCLD, SIG_IGN);   /* no wait() */
#else
  signal(SIGCHLD, SIG_IGN); 
#endif
  signal(SIGTERM, closedown);
  signal(SIGSTOP, closedown);
  signal(SIGCONT, closedown);
  (void)write_my_pid();
  if (setuid(TFTP_UID) == -1) 
    fprintf(logfile,"\n*** WARNING : starting up as Super User\n"); 
  /*
   *
   *  Daemon loop.
   *
   */
  {
  int addrlen=sizeof(struct sockaddr_in), recv_cnt, today;
  long tloc;
  struct tm *tms;
    /* time stamp start up */
    time(&tloc);  
    tms=localtime(&tloc);
    today=tms->tm_mday;
    fprintf(logfile,"\n*** STARTING\n\n"); 
    fprintf(logfile,"*** TIME: %s",ctime(&tloc)); 
    for (;;)
    {
#ifdef DEBUG
      fprintf(logfile,"Waiting for client request ... \n");
#endif
      /* wait for and read client request */
      if ((recv_cnt=recvfrom(lis, buf, sizeof(buf), 0, &from, &addrlen)) == -1) 
      {
        if (errno == EINTR) continue;   /* caught signal */
        fprintf(logfile,"Unable to accept call on listen socket: %s\n",sys_errlist[errno]);
        sleep(RETRY);
        continue;
      }
      /* flag first call of the day */
      time(&tloc);  
      tms=localtime(&tloc);
      if (tms->tm_mday != today)
      {
        fprintf(logfile,"*** %s ",ctime(&tloc)); 
        today=tms->tm_mday;
      }
      /* log client */
      ++rnum;   /* log number of call */
      if ((hp=gethostbyaddr((char *) &from.sin_addr, sizeof(struct in_addr), from.sin_family)) == NULL)
        fprintf(logfile,"*** call: %04u client: %s TIME: %02d:%02d:%02d\n",rnum,inet_ntoa(from.sin_addr),tms->tm_hour,tms->tm_min,tms->tm_sec);
      else
        fprintf(logfile,"*** call: %04u client: %s TIME: %02d:%02d:%02d\n",rnum,hp->h_name,tms->tm_hour,tms->tm_min,tms->tm_sec);
      /* spawn client server */
      switch(fork())
      {
      case -1:
               fprintf(logfile,"*** call: %04u could not fork server: %s\n",sys_errlist[errno]);
               break;   
      case 0:  close(lis);
               serve_request(1, recv_cnt);
               exit(0);
      default: break;
      } 
    }    
  }  
} /* main */

/*
 *  Log shutdowns.
 */
void closedown(sig)
int sig;
{
long tloc;
struct tm *tms;
  time(&tloc); 
  tms=localtime(&tloc);
  switch(sig)
  {
  case SIGCONT:
               fprintf(logfile,"Re-started: %s TIME: %02d:%02d:%02d\n",tms->tm_hour,tms->tm_min,tms->tm_sec);
               break;
  case SIGSTOP:
               fprintf(logfile,"Stopped ... %s TIME: %02d:%02d:%02d\n",tms->tm_hour,tms->tm_min,tms->tm_sec);
               break;
  default:
               fprintf(logfile,"Going down on signal %d TIME: %02d:%02d:%02d\n",sig,tms->tm_hour,tms->tm_min,tms->tm_sec);
               exit(0);
  }
}

void help()
{
  puts("\nusage : tftpd [-k] | [-l] [-s] dir1 dir2 .. dir4\n");
  puts("         -k : terminate the running tftp daemon\n");
  puts("         -l : log daemon connections and errors\n");
  puts("         -s : start tftp as daemon\n");
  exit(0);
}

/*
 *  Record Pid of running daemon into
 *  a file.
 */
void write_my_pid()
{
int fd;
char wbuf[5];
pid_t pid=9999;
  pid=getpid();
  sprintf(wbuf,"%d\n",pid);
  if ((fd=open(PID, O_CREAT | O_WRONLY, 0644)) == -1)
  {
    fprintf(stderr,"Could not open %s: %s\n",PID,sys_errlist[errno]);
    exit(1);
  }
  if (write(fd, wbuf, 6) == -1)
  {
    fprintf(stderr,"Could not write to %s: %s\n",PID,sys_errlist[errno]);
    exit(1);
  } 
}

/*
 *  Send control signal to running
 *  daemon.
 */
void signal_current_tftpd(sig)
int sig;
{
int fd;
char rbuf[5];
pid_t pid=9999;
  if ((fd=open(PID, O_RDONLY)) == -1)
  {
    fprintf(logfile,"*** Could not open %s: %s\n",PID,sys_errlist[errno]);
    exit(1);
  }
  if (read(fd, rbuf, 5) == -1)
  {
    fprintf(logfile,"*** Could not read %s: %s\n",PID,sys_errlist[errno]);
    exit(1);
  }
  if ((pid=atoi(rbuf)) == 0)
  {
    fprintf(logfile,"*** Bad PID read from pid file: %s\n",rbuf);
    exit(1);
  }
  if (kill(pid, sig) == -1)
  {
    fprintf(logfile,"*** Error sending signal %d to PID %d: %s\n",sig,pid,sys_errlist[errno]);
    exit(1);
  }
}
